"""Implementation of the FedDyn [FedDyn21]_ algorithm.

References:
    .. [FedDyn21] Durmus Alp Emre Acar, Yue Zhao, Ramon Matas, Matthew Mattina, Paul Whatmough,
       and Venkatesh Saligrama. Federated Learning with Dynamic Regularization.
       In ICLR (2021). URL: https://openreview.net/pdf?id=B7v4QMR6Z9w
"""

import sys
from copy import deepcopy
from typing import Collection, Sequence

import numpy as np
import torch
from torch.nn import Module

sys.path.append(".")
sys.path.append("..")

from .. import FlukeENV  # NOQA
from ..client import Client  # NOQA
from ..comm import Message  # NOQA
from ..config import OptimizerConfigurator  # NOQA
from ..data import FastDataLoader  # NOQA
from ..server import Server  # NOQA
from ..utils import clear_cuda_cache  # NOQA
from ..utils.model import aggregate_models, safe_load_state_dict  # NOQA
from . import CentralizedFL  # NOQA

__all__ = ["FedDynClient", "FedDynServer", "FedDyn"]


def get_all_params_of(model: Module, copy: bool = True) -> torch.Tensor:
    result = None
    for param in model.parameters():
        if not param.requires_grad:
            continue

        if result is None:
            result = param.clone().detach().reshape(-1) if copy else param.reshape(-1)
        else:
            result = (
                torch.cat((result, param.clone().detach().reshape(-1)), 0)
                if copy
                else torch.cat((result, param.reshape(-1)), 0)
            )
    return result


def load_all_params(device: torch.device, model: torch.nn.Module, params: torch.Tensor) -> None:
    dict_param = deepcopy(dict(model.named_parameters()))
    idx = 0
    for name, param in model.named_parameters():
        if not param.requires_grad:
            continue
        weights = param.data
        length = len(weights.reshape(-1))
        dict_param[name].data.copy_(
            params[idx : idx + length].clone().detach().reshape(weights.shape).to(device)
        )
        idx += length

    model.load_state_dict(dict_param, strict=False)


class FedDynClient(Client):

    def __init__(
        self,
        index: int,
        train_set: FastDataLoader,
        test_set: FastDataLoader,
        optimizer_cfg: OptimizerConfigurator,
        loss_fn: torch.nn.Module,
        local_epochs: int,
        alpha: float,
        fine_tuning_epochs: int = 0,
        clipping: float = 5,
        **kwargs,
    ):
        super().__init__(
            index=index,
            train_set=train_set,
            test_set=test_set,
            optimizer_cfg=optimizer_cfg,
            loss_fn=loss_fn,
            local_epochs=local_epochs,
            fine_tuning_epochs=fine_tuning_epochs,
            clipping=clipping,
            **kwargs,
        )

        self.hyper_params.update(alpha=alpha)
        self.weight = None
        self.weight_decay = (
            self._optimizer_cfg.optimizer_cfg["weight_decay"]
            if "weight_decay" in self._optimizer_cfg.optimizer_cfg
            else 0
        )
        self.prev_grads = None
        self._attr_to_cache.extend(["prev_grads", "weight"])

    def receive_model(self) -> None:
        model, cld_mdl = self.channel.receive(self.index, "server", msg_type="model").payload
        if self.model is None:
            self.model = model
            self.prev_grads = torch.zeros_like(get_all_params_of(self.model), device=self.device)
        else:
            safe_load_state_dict(self.model, cld_mdl.state_dict())

    def _receive_weights(self) -> None:
        self.weight = self.channel.receive(self.index, "server", msg_type="weight").payload

    def _send_weight(self) -> None:
        self.channel.send(
            Message(self.train_set.tensors[0].shape[0], "weight", self.index, inmemory=True),
            "server",
        )

    def send_model(self) -> None:
        self.channel.send(Message(self.model, "model", self.index, inmemory=True), "server")
        self.channel.send(Message(self.prev_grads, "grads", self.index, inmemory=True), "server")

    def fit(self, override_local_epochs: int = 0) -> float:
        epochs: int = (
            override_local_epochs if override_local_epochs > 0 else self.hyper_params.local_epochs
        )

        self.model.train()
        self.model.to(self.device)

        alpha_coef_adpt = self.hyper_params.alpha / self.weight
        server_params = get_all_params_of(self.model).to(self.device)

        for params in self.model.parameters():
            params.requires_grad = True

        if not self.optimizer:
            w_dec = alpha_coef_adpt + self.weight_decay
            self.optimizer, self.scheduler = self._optimizer_cfg(
                self.model,
                # this override the weight_decay
                # in the optimizer_cfg
                weight_decay=w_dec,
            )
        running_loss = 0.0
        for _ in range(epochs):
            for _, (X, y) in enumerate(self.train_set):
                X, y = X.to(self.device), y.to(self.device)
                self.optimizer.zero_grad()
                y_hat = self.model(X)
                loss = self.hyper_params.loss_fn(y_hat, y)

                # Dynamic regularization
                curr_params = get_all_params_of(self.model, False).to(self.device)
                # penalty = -torch.sum(curr_params * self.prev_grads)
                # penalty += 0.5 * alpha_coef_adpt * torch.sum((curr_params - server_params) ** 2)
                penalty = alpha_coef_adpt * torch.sum(
                    curr_params * (-server_params + self.prev_grads)
                )
                loss = loss + penalty

                loss.backward()
                self._clip_grads(self.model)
                self.optimizer.step()
                running_loss += loss.item()

            self.scheduler.step()

        # update the previous gradients
        curr_params = get_all_params_of(self.model).to(self.device)
        self.prev_grads += alpha_coef_adpt * (server_params - curr_params)

        running_loss /= epochs * len(self.train_set)
        self.model.cpu()
        clear_cuda_cache()
        return running_loss


class FedDynServer(Server):
    def __init__(
        self,
        model: Module,
        test_set: FastDataLoader,
        clients: Sequence[Client],
        weighted: bool = True,
        alpha: float = 0.01,
    ):
        super().__init__(model=model, test_set=test_set, clients=clients, weighted=weighted)
        self.alpha = alpha
        self.device = FlukeENV().get_device()
        self.cld_mdl = deepcopy(self.model).to(self.device)

    def broadcast_model(self, eligible: Sequence[Client]) -> None:
        self.channel.broadcast(
            Message((self.model, self.cld_mdl), "model", "server"), [c.index for c in eligible]
        )

    def fit(
            self, n_rounds: int = 10, eligible_perc: float = 0.1, finalize: bool = True, **kwargs
    ) -> None:

        # Weight computation
        for client in self.clients:
            client._send_weight()

        weights = np.array(
            [
                self.channel.receive("server", client.index, msg_type="weight").payload
                for client in self.clients
            ]
        )
        weights = weights / np.sum(weights) * self.n_clients

        for i, client in enumerate(self.clients):
            self.channel.send(Message(weights[i], "weight", "server"), client.index)

        for client in self.clients:
            client._receive_weights()

        return super().fit(n_rounds, eligible_perc)

    @torch.no_grad()
    def aggregate(self, eligible: Sequence[Client], client_models: Collection[Module]) -> None:
        weights = self._get_client_weights(eligible)
        agg_model_sd = aggregate_models(
            self.model, client_models, weights, eta=self.hyper_params.lr, inplace=False
        ).state_dict()

        avg_grad = None
        grad_count = 0
        for client in eligible:
            pg = self.channel.receive("server", client.index, msg_type="grads").payload
            if pg is not None:
                grad_count += 1
                if avg_grad is None:
                    avg_grad = pg.clone().detach()
                else:
                    avg_grad += pg.clone().detach()

        if grad_count > 0:
            avg_grad /= grad_count

        self.model.load_state_dict(agg_model_sd)
        load_all_params(
            self.device,
            self.cld_mdl,
            get_all_params_of(self.model).to(self.device) + avg_grad.to(self.device),
        )


class FedDyn(CentralizedFL):

    def get_client_class(self) -> type[Client]:
        return FedDynClient

    def get_server_class(self) -> type[Server]:
        return FedDynServer



"""This module contains classes and functions for parallel federated learning using multiple
GPUs."""

__all__ = ["client", "server", "ParallelAlgorithm"]


from copy import deepcopy
from typing import Any, Sequence

from .. import DDict  # NOQA
from ..algorithms import CentralizedFL  # NOQA
from ..config import OptimizerConfigurator  # NOQA
from ..data import DataSplitter  # NOQA
from ..data import FastDataLoader  # NOQA
from ..utils import get_loss  # NOQA
from .client import ParallelClient  # NOQA
from .server import ParallelServer  # NOQA
from .utils import ModelBuilder  # NOQA


class ParallelAlgorithm(CentralizedFL):
    """
    ParallelAlgorithm is a class for parallel federated learning algorithms that use multiple
    GPUs. It inherits from CentralizedFL and initializes clients and server for parallel execution.
    Clients training runs on different GPUs, and the server aggregates their updates.
    """

    def __init__(
        self,
        n_clients: int,
        data_splitter: DataSplitter,
        hyper_params: DDict | dict[str, Any],
        clients: list[ParallelClient] = None,
        server: ParallelServer = None,
        **kwargs,
    ):
        self._builder = ModelBuilder(
            hyper_params.model, **hyper_params.model_args if "model_args" in hyper_params else {}
        )
        super().__init__(
            n_clients=n_clients,
            data_splitter=data_splitter,
            hyper_params=hyper_params,
            clients=clients,
            server=server,
            **kwargs,
        )

    def init_clients(
        self,
        clients_tr_data: list[FastDataLoader],
        clients_te_data: list[FastDataLoader],
        config: DDict,
    ) -> Sequence[ParallelClient]:

        self._fix_opt_cfg(config.optimizer)
        optimizer_cfg = OptimizerConfigurator(
            optimizer_cfg=config.optimizer, scheduler_cfg=config.scheduler
        )
        loss = get_loss(config.loss) if isinstance(config.loss, str) else config.loss()
        clients = [
            self.get_client_class()(
                builder=self._builder,
                index=i,
                train_set=clients_tr_data[i],
                test_set=clients_te_data[i],
                optimizer_cfg=optimizer_cfg,
                loss_fn=deepcopy(loss),
                **config.exclude("optimizer", "loss", "batch_size", "scheduler"),
            )
            for i in range(self.n_clients)
        ]
        return clients

    def get_server_class(self) -> type[ParallelServer]:
        return ParallelServer

    def get_client_class(self) -> type[ParallelClient]:
        return ParallelClient



import sys
import time
from copy import deepcopy
from queue import SimpleQueue
from typing import Any, Iterable

import torch.multiprocessing as mp

sys.path.append(".")
sys.path.append("..")

from .. import FlukeENV  # NOQA
from ..config import ConfigurationError  # NOQA
from ..data import FastDataLoader  # NOQA
from ..evaluation import Evaluator  # NOQA
from ..server import EarlyStopping, Server  # NOQA
from .client import ParallelClient  # NOQA

__all__ = ["ParallelServer"]


def _train_worker(
    client: ParallelClient,
    model: dict,
    round: int,
    device: str,
    prefit: bool,
    postfit: bool,
    evaluator: Evaluator,
    result_dict: dict,
) -> None:
    result_dict.put(
        client.local_update(
            current_round=round,
            current_model_sd=model,
            device=device,
            prefit=prefit,
            postfit=postfit,
            evaluator=evaluator,
        )
    )


class ParallelServer(Server):
    """Server for parallel federated learning using multiple GPUs.

    This class is a subclass of :class:`Server` that allows the use of multiple GPUs - one
    for each client. It uses multiprocessing to handle the training of clients in parallel.
    It is designed to work with the `torch.multiprocessing` module, which allows for
    parallel execution of client training processes across multiple GPUs.
    """

    def __init__(self, *args: Any, **kwargs: dict[str, Any]) -> None:
        super().__init__(*args, **kwargs)

        if len(FlukeENV().get_device_ids()) <= 1:
            raise ConfigurationError(
                "ParallelServer requires at least 2 GPUs."
                + "Please set a list of devices in the config file."
            )
        self.device = FlukeENV().get_device()
        mp.set_start_method("spawn")
        self.manager = mp.Manager()
        self.result_dict = self.manager.dict()
        self.processes = []
        self.running_processes = []

    def _clients_local_fit(self, eligible: Iterable[ParallelClient]) -> dict[int, dict]:
        console = FlukeENV().get_progress_bar("clients").console
        trainer_queue = SimpleQueue()
        for client in eligible:
            trainer_queue.put(client)

        self.available_gpus = [f"cuda:{i}" for i in FlukeENV().get_device_ids()]
        result_queue = self.manager.Queue()
        current_model = deepcopy(self.model.state_dict())
        current_round = self.rounds + 1
        result_dict = {}
        while not trainer_queue.empty() or self.running_processes:
            while self.available_gpus and not trainer_queue.empty():
                device = self.available_gpus.pop(0)
                client = trainer_queue.get()
                p = mp.Process(
                    target=_train_worker,
                    args=(
                        client,
                        current_model,
                        current_round,
                        device,
                        FlukeENV().get_eval_cfg().pre_fit,
                        FlukeENV().get_eval_cfg().post_fit,
                        FlukeENV().get_evaluator(),
                        result_queue,
                    ),
                )
                p.start()
                self.running_processes.append((p, device))

            # Check if any process finished
            for p, gpu_id in self.running_processes.copy():
                if not p.is_alive():
                    p.join()
                    self.running_processes.remove((p, gpu_id))
                    self.available_gpus.append(gpu_id)

            # Process results if any
            while not result_queue.empty():
                model, loss, idx, results = result_queue.get()
                result_dict[idx] = (model, results)
                console.log(f"Client {idx} local update complete [loss={loss}]")
                # console.log(f"Client {idx} evaluation \n{results}")

            time.sleep(1)  # Avoid busy waiting

        return result_dict

    def evaluate(self, evaluator: Evaluator, test_set: FastDataLoader) -> dict[str, float]:
        if test_set is not None:
            return evaluator.evaluate(self.rounds + 1, self.model, test_set, device=self.device[0])
        return {}

    def fit(
        self, n_rounds: int = 10, eligible_perc: float = 0.1, finalize: bool = True, **kwargs
    ) -> None:

        total_rounds = self.rounds + n_rounds
        for rnd in range(self.rounds, total_rounds):
            try:
                self.notify(event="start_round", round=rnd + 1, global_model=self.model)
                eligible = self.get_eligible_clients(eligible_perc)
                self.notify(event="selected_clients", round=rnd + 1, clients=eligible)
                client_modopts = self._clients_local_fit(eligible)
                for client in eligible:
                    self._participants.add(client.index)
                    c_modopt, c_results = client_modopts[client.index]
                    self.clients[client.index]._modopt = c_modopt

                    for k, evals in c_results:
                        self._notify(
                            "client_evaluation",
                            round=rnd + 1,
                            client_id=client.index,
                            phase=k,
                            results=evals,
                        )

                client_models = [self.clients[c.index].model for c in eligible]
                self.aggregate(eligible, client_models)
                self._compute_evaluation(rnd, eligible)
                self.notify(event="end_round", round=rnd + 1)
                self.rounds += 1

            except KeyboardInterrupt:
                self.notify(event="interrupted")
                break

            except EarlyStopping:
                self.notify(event="early_stop")
                break

        if finalize:
            self.finalize()



from __future__ import annotations

import sys

import pytest

sys.path.append(".")
sys.path.append("..")


import fluke.run  # noqa: E402


def test_federation():
    fluke.run.federation(
        exp_cfg="./tests/configs/exp.yaml",
        alg_cfg="./tests/configs/alg/fedavg.yaml",
        overrides=["exp.seed=98765"],
        resume=None,
    )


def test_clients_only():
    fluke.run.clients_only(
        exp_cfg="./tests/configs/exp.yaml", alg_cfg="./tests/configs/alg/fedavg.yaml", epochs=1
    )


def test_centralized():
    fluke.run.centralized(
        exp_cfg="./tests/configs/exp.yaml", alg_cfg="./tests/configs/alg/fedavg.yaml", epochs=1
    )


def test_sweep():
    fluke.run.sweep(
        exp_cfg="./tests/configs/exp.yaml",
        alg_cfgs=["./tests/configs/alg/fedavg.yaml", "./tests/configs/alg/fedprox.yaml"],
    )


def test_run_others():
    fluke.run.run(True)
    with pytest.raises(Exception):
        fluke.run.version_callback(True)


if __name__ == "__main__":
    test_federation()
    test_clients_only()
    test_centralized()



"""Implementation of the [FedRS21]_ algorithm.

References:
    .. [FedRS21] Xin-Chun Li and De-Chuan Zhan. FedRS: Federated Learning with Restricted
       Softmax for Label Distribution Non-IID Data. In KDD (2021).
       URL: https://doi.org/10.1145/3447548.3467254
"""

import sys
from typing import Literal

import numpy as np
import torch

sys.path.append(".")
sys.path.append("..")

from ..client import Client  # NOQA
from ..config import OptimizerConfigurator  # NOQA
from ..data import FastDataLoader  # NOQA
from . import CentralizedFL  # NOQA

__all__ = ["RSLoss", "FedRSClient", "FedRS"]


class RSLoss(torch.nn.Module):
    """Restricted Softmax Loss function.

    Args:
        class_scaling (torch.Tensor): Class scaling factor.
        reduction (Literal["mean", "sum"]): Specifies the reduction to apply to the output.

    See Also:
        This loss function is very similar to the one used in ``FedLC``, i.e.,
        :class:`fluke.algorithms.fedlc.CalibratedLoss`.
    """

    def __init__(self, class_scaling: torch.Tensor, reduction: Literal["mean", "sum"] = "mean"):
        super().__init__()
        self.class_scaling = class_scaling
        self.reduction = reduction

    def forward(self, logit: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        out = logit * self.class_scaling.to(logit.device)
        return torch.nn.functional.cross_entropy(out, y, reduction=self.reduction)

    def __str__(self, indent: int = 0) -> str:
        indent_str = " " * indent
        return f"{indent_str}RSLoss()"

    def __repr__(self, indent: int = 0) -> str:
        return self.__str__(indent=indent)


class FedRSClient(Client):

    def __init__(
        self,
        index: int,
        train_set: FastDataLoader,
        test_set: FastDataLoader,
        optimizer_cfg: OptimizerConfigurator,
        loss_fn: torch.nn.Module,  # ignored
        local_epochs: int,
        alpha: float,
        count_as_missing: int = 2,
        fine_tuning_epochs: int = 0,
        clipping: float = 0,
        **kwargs,
    ):
        uniq_val, uniq_cnt = np.unique(train_set.tensors[1], return_counts=True)
        class_scaling = torch.ones(train_set.num_labels) * alpha
        for i, c in enumerate(uniq_val):
            if uniq_cnt[i] > count_as_missing:
                class_scaling[c] = 1.0
        class_scaling = class_scaling.unsqueeze(dim=0)
        super().__init__(
            index=index,
            train_set=train_set,
            test_set=test_set,
            optimizer_cfg=optimizer_cfg,
            loss_fn=RSLoss(class_scaling),
            local_epochs=local_epochs,
            fine_tuning_epochs=fine_tuning_epochs,
            clipping=clipping,
            **kwargs,
        )
        self.hyper_params.update(alpha=alpha, count_as_missing=count_as_missing)


class FedRS(CentralizedFL):

    def get_client_class(self) -> type[Client]:
        return FedRSClient



from __future__ import annotations

import sys

sys.path.append(".")
sys.path.append("..")


import fluke.get  # noqa: E402


def test_get():

    fluke.get.list()
    fluke.get.config(name="exp", outdir="./tmp")
    fluke.get.config(name="fedavg", outdir="./tmp")


if __name__ == "__main__":
    test_get()



"""Implementation of the [FedProx18]_ algorithm.

References:
    .. [FedProx18] Tian Li, Anit Kumar Sahu, Manzil Zaheer, Maziar Sanjabi, Ameet Talwalkar,
       and Virginia Smith. Federated Optimization in Heterogeneous Networks. Adaptive & Multitask
       Learning Workshop. URL: https://openreview.net/pdf?id=SkgwE5Ss3N
"""

import sys
from copy import deepcopy

import torch

sys.path.append(".")
sys.path.append("..")

from ..client import Client  # NOQA
from ..config import OptimizerConfigurator  # NOQA
from ..data import FastDataLoader  # NOQA
from ..utils import clear_cuda_cache  # NOQA
from . import CentralizedFL  # NOQA

__all__ = ["FedProxClient", "FedProx"]


class FedProxClient(Client):
    def __init__(
        self,
        index: int,
        train_set: FastDataLoader,
        test_set: FastDataLoader,
        optimizer_cfg: OptimizerConfigurator,
        loss_fn: torch.nn.Module,
        local_epochs: int,
        mu: float,
        fine_tuning_epochs: int = 0,
        clipping: float = 0,
        **kwargs,
    ):
        super().__init__(
            index=index,
            train_set=train_set,
            test_set=test_set,
            optimizer_cfg=optimizer_cfg,
            loss_fn=loss_fn,
            local_epochs=local_epochs,
            fine_tuning_epochs=fine_tuning_epochs,
            clipping=clipping,
            **kwargs,
        )
        self.hyper_params.update(mu=mu)

    def _proximal_loss(self, local_model: torch.nn.Module, global_model: torch.nn.Module) -> float:
        proximal_term = 0.0
        for w, w_t in zip(local_model.parameters(), global_model.parameters()):
            if w_t.requires_grad:
                proximal_term += (w - w_t).norm(2) ** 2
        return proximal_term

    def fit(self, override_local_epochs: int = 0) -> float:
        epochs: int = (
            override_local_epochs if override_local_epochs > 0 else self.hyper_params.local_epochs
        )
        W = deepcopy(self.model).to(self.device)
        self.model.to(self.device)
        self.model.train()

        if self.optimizer is None:
            self.optimizer, self.scheduler = self._optimizer_cfg(self.model)

        running_loss = 0.0
        for _ in range(epochs):
            loss = None
            for _, (X, y) in enumerate(self.train_set):
                X, y = X.to(self.device), y.to(self.device)
                self.optimizer.zero_grad()
                y_hat = self.model(X)
                loss = self.hyper_params.loss_fn(y_hat, y) + (
                    self.hyper_params.mu / 2
                ) * self._proximal_loss(self.model, W)
                loss.backward()
                self._clip_grads(self.model)
                self.optimizer.step()
                running_loss += loss.item()
            self.scheduler.step()

        running_loss /= epochs * len(self.train_set)
        self.model.cpu()
        W.cpu()
        clear_cuda_cache()
        return running_loss


class FedProx(CentralizedFL):

    def get_client_class(self) -> type[Client]:
        return FedProxClient



"""Implementation of the DPFedAVG: Differential Privacy Federated Averaging [DPFedAVG2017]_
algorithm.

Note:
    This implementation does not exactly follow the original paper, but it is a simplified version
    that uses the Opacus library to provide differential privacy guarantees.

References:
    .. [DPFedAVG2017] Robin C. Geyer, Tassilo Klein, Moin Nabi.
       Differentially Private Federated Learning: A Client Level Perspective
       In ArXiv (2017). URL: https://arxiv.org/abs/1712.07557
"""

import sys
from typing import Sequence

import torch
from opacus import PrivacyEngine
from torch.nn import Module

sys.path.append(".")
sys.path.append("..")

from ..client import Client  # NOQA
from ..config import OptimizerConfigurator  # NOQA
from ..data import FastDataLoader  # NOQA
from ..server import Server  # NOQA
from . import CentralizedFL  # NOQA

__all__ = ["DPFedAVG", "DPFedAVGClient", "DPFedAVGServer"]


class _OpacusModelAdapter(Module):
    """Adapt a model to be compatible with Opacus.
    Opacus encapsultes the model in a way that the parameter names have a prefix
    named "_module". This class is a simple adapter to make the model compatible
    with Opacus.

    Args:
        model (Module): The model to be adapted.
    """

    def __init__(self, model: Module):
        super().__init__()
        self._module = model

    def forward(self, *args, **kwargs) -> torch.Tensor:
        return self._module(*args, **kwargs)


class DPFedAVGClient(Client):
    def __init__(
        self,
        index: int,
        train_set: FastDataLoader,
        test_set: FastDataLoader,
        optimizer_cfg: OptimizerConfigurator,
        loss_fn: Module,
        local_epochs: int = 3,
        fine_tuning_epochs: int = 0,
        clipping: float = 0,
        noise_mul: float = 1.1,
        max_grad_norm: float = 1.0,
        **kwargs,
    ):
        super().__init__(
            index=index,
            train_set=train_set,
            test_set=test_set,
            optimizer_cfg=optimizer_cfg,
            loss_fn=loss_fn,
            local_epochs=local_epochs,
            fine_tuning_epochs=fine_tuning_epochs,
            clipping=clipping,
            **kwargs,
        )
        self.hyper_params.update(noise_mul=noise_mul, max_grad_norm=max_grad_norm)

    def _init_private_engine(self) -> None:
        if self.model is None:
            return
        self.privacy_engine = PrivacyEngine()
        self.model.train()
        self.model, self.optimizer, self.train_set = self.privacy_engine.make_private(
            module=self.model._module,
            optimizer=self.optimizer,
            data_loader=(
                self.train_set.as_dataloader()
                if isinstance(self.train_set, FastDataLoader)
                else self.train_set
            ),
            noise_multiplier=self.hyper_params.noise_mul,
            max_grad_norm=self.hyper_params.max_grad_norm,
        )

    def receive_model(self) -> None:
        if self.model is None:
            super().receive_model()
            self.optimizer, self.scheduler = self._optimizer_cfg(self.model)
            return self._init_private_engine()
        else:
            return super().receive_model()


class DPFedAVGServer(Server):

    def __init__(
        self,
        model: Module,
        test_set: FastDataLoader,
        clients: Sequence[Client],
        weighted: bool = False,
        lr: float = 1.0,
        **kwargs,
    ):
        super().__init__(
            model=_OpacusModelAdapter(model),
            test_set=test_set,
            clients=clients,
            weighted=weighted,
            lr=lr,
            **kwargs,
        )


class DPFedAVG(CentralizedFL):

    def get_server_class(self) -> type[Server]:
        return DPFedAVGServer

    def get_client_class(self) -> type[Client]:
        return DPFedAVGClient
